"""Add label YAML."""

from __future__ import annotations

import json
from functools import partial
from typing import Any

from omegaconf import OmegaConf

import gdsfactory as gf
from gdsfactory.serialization import clean_dict
from gdsfactory.typings import LayerSpec


@gf.cell_with_child
def add_label_yaml(
    component: gf.Component,
    layer: LayerSpec = "TEXT",
    measurement: str | None = None,
    measurement_settings: dict[str, Any] | None = None,
    analysis: str | None = None,
    analysis_settings: dict[str, Any] | None = None,
    doe: str | None = None,
    with_yaml_format: bool = True,
    port_index_optical: tuple[int, ...] | None = (0,),
    port_index_electrical: tuple[int, ...] | None = (0,),
) -> gf.Component:
    """Returns new Component with measurement label.

    Args:
        component: to add labels to.
        layer: text label layer.
        measurement: measurement config name. Defaults to component['info']['measurement'].
        measurement_settings: measurement settings. Defaults to component['info']['measurement_settings'].
        analysis: analysis name. Defaults to component['info']['analysis'].
        analysis_settings: Extra analysis settings. Defaults to component settings.
        doe: Design of Experiment name. Defaults to component['info']['doe'].
        with_yaml_format: whether to use yaml or json format.
        port_index_optical: port index to add to the label. None adds it to all, 0 to first, -1 to last.
        port_index_electrical: port index to add to the label. None adds it to all, 0 to first, -1 to last.
    """
    from gdsfactory.pdk import get_layer

    c = gf.Component()
    ref = c << gf.get_component(component)
    c.add_ports(ref.ports)

    measurement = measurement or component.info.get("measurement")
    measurement_settings = measurement_settings or component.info.get(
        "measurement_settings"
    )
    analysis = analysis or component.info.get("analysis")
    analysis_settings = analysis_settings or component.info.get("analysis_settings")
    doe = doe or component.info.get("doe")

    layer = get_layer(layer)
    analysis_settings = analysis_settings or {}
    measurement_settings = measurement_settings or {}
    cell_settings = dict(component.settings)
    cell_settings.update(dict(component.info))
    cell_settings = clean_dict(cell_settings)

    optical_ports = component.get_ports_list(port_type="optical")
    electrical_ports = component.get_ports_list(port_type="electrical")

    port_index_optical = port_index_optical if optical_ports else ()
    port_index_electrical = port_index_electrical if electrical_ports else ()

    port_names_optical = [p.name for p in optical_ports]
    port_names_electrical = [p.name for p in electrical_ports]

    settings = dict(
        name=component.name,
        doe=doe,
        measurement=measurement,
        cell_settings=cell_settings,
        analysis=analysis,
        measurement_settings=measurement_settings,
        analysis_settings=analysis_settings,
    )
    for port_index in port_index_optical:
        d = dict(port_type="optical", port_names=port_names_optical, **settings)
        text = OmegaConf.to_yaml(d) if with_yaml_format else json.dumps(d)
        x, y = optical_ports[port_index].center
        label = gf.Label(
            text=text,
            origin=(x, y),
            anchor="o",
            layer=layer[0],
            texttype=layer[1],
        )
        c.add(label)
    for port_index in port_index_electrical:
        d = dict(port_type="electrical", port_names=port_names_electrical, **settings)
        text = OmegaConf.to_yaml(d) if with_yaml_format else json.dumps(d)
        x, y = electrical_ports[port_index].center
        label = gf.Label(
            text=text,
            origin=(x, y),
            anchor="o",
            layer=layer[0],
            texttype=layer[1],
        )
        c.add(label)
    c.copy_child_info(component)
    return c


add_label_json = partial(add_label_yaml, with_yaml_format=False)


if __name__ == "__main__":
    measurement_settings = dict(
        wavelenth_min=1550, wavelenth_max=1570, wavelength_steps=10
    )
    with_yaml_format = False
    with_yaml_format = True

    info = dict(
        measurement_settings=measurement_settings,
        with_yaml_format=with_yaml_format,
    )

    c = gf.c.straight(length=11)
    c = gf.c.mmi2x2(length_mmi=2.2)
    c = gf.routing.add_fiber_array(
        c,
        get_input_labels_function=None,
        grating_coupler=gf.components.grating_coupler_te,
    )
    c = add_label_json(c)
    info = dict(
        measurement="optical_loopback2",
        doe="spiral_sc",
        wavelenth_min=1560,
    )
    c.info.update(info)

    # c = gf.components.spiral_inner_io_fiber_array(
    #     length=20e3,
    #     decorator=decorator,
    #     info=dict(
    #         measurement="optical_loopback2",
    #         doe="spiral_sc",
    #         measurement_settings=dict(wavelength_alignment=1560),
    #     ),
    # )
    # print(len(c.labels[0].text))
    # print(c.labels[0].text)
    # d = yaml.safe_load(c.labels[0].text) if yaml else json.loads(c.labels[0].text)
    # print(d)
    c.show(show_ports=False)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.typings import ComponentFactory, CrossSectionSpec


@gf.cell
def mmi2x2(
    width: float | None = None,
    width_taper: float = 1.0,
    length_taper: float = 10.0,
    length_mmi: float = 5.5,
    width_mmi: float = 2.5,
    gap_mmi: float = 0.25,
    taper: ComponentFactory = taper_function,
    cross_section: CrossSectionSpec = "xs_sc",
    **kwargs,
) -> Component:
    r"""Mmi 2x2.

    Args:
        width: input and output straight width.
        width_taper: interface between input straights and mmi region.
        length_taper: into the mmi region.
        length_mmi: in x direction.
        width_mmi: in y direction.
        gap_mmi: (width_taper + gap between tapered wg)/2.
        taper: taper function.
        straight: straight function.
        with_bbox: add rectangular box in cross_section
            bbox_layers and bbox_offsets to avoid DRC sharp edges.
        cross_section: spec.


    .. code::

                   length_mmi
                    <------>
                    ________
                   |        |
                __/          \__
            o2  __            __  o3
                  \          /_ _ _ _
                  |         | _ _ _ _| gap_mmi
                __/          \__
            o1  __            __  o4
                  \          /
                   |________|

                 <->
            length_taper

    """
    c = gf.Component()
    gap_mmi = gf.snap.snap_to_grid(gap_mmi, grid_factor=2)
    w_taper = width_taper
    x = gf.get_cross_section(cross_section, **kwargs)
    width = width or x.width

    _taper = taper(
        length=length_taper,
        width1=width,
        width2=w_taper,
        cross_section=x,
    )

    delta_width = width_mmi - width
    y = width_mmi / 2
    c.add_polygon([(0, -y), (length_mmi, -y), (length_mmi, y), (0, y)], layer=x.layer)
    for section in x.sections[1:]:
        layer = section.layer
        y = section.width / 2 + delta_width / 2
        c.add_polygon(
            [
                (-delta_width, -y),
                (length_mmi + delta_width, -y),
                (length_mmi + delta_width, y),
                (-delta_width, y),
            ],
            layer=layer,
        )

    a = gap_mmi / 2 + width_taper / 2
    ports = [
        gf.Port("o1", orientation=180, center=(0, -a), width=w_taper, cross_section=x),
        gf.Port("o2", orientation=180, center=(0, +a), width=w_taper, cross_section=x),
        gf.Port(
            "o3",
            orientation=0,
            center=(length_mmi, +a),
            width=w_taper,
            cross_section=x,
        ),
        gf.Port(
            "o4",
            orientation=0,
            center=(length_mmi, -a),
            width=w_taper,
            cross_section=x,
        ),
    ]

    for port in ports:
        taper_ref = c << _taper
        taper_ref.connect(port="o2", destination=port)
        c.add_port(name=port.name, port=taper_ref.ports["o1"])
        c.absorb(taper_ref)

    x.add_bbox(c)
    return c


if __name__ == "__main__":
    # c = mmi2x2(gap_mmi=0.252, cross_section="xs_m1")
    c = mmi2x2(gap_mmi=0.252, cross_section="xs_rc")
    c.show(show_ports=True)
    c.pprint()



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.coupler_ring import coupler_ring as coupler_ring_function
from gdsfactory.components.straight import straight
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@gf.cell
def ring_double(
    gap: float = 0.2,
    gap_top: float | None = None,
    radius: float = 10.0,
    length_x: float = 0.01,
    length_y: float = 0.01,
    coupler_ring: ComponentSpec = coupler_ring_function,
    bend: ComponentSpec = bend_euler,
    straight: ComponentSpec = straight,
    cross_section: CrossSectionSpec = "xs_sc",
) -> Component:
    """Returns a double bus ring.

    two couplers (ct: top, cb: bottom)
    connected with two vertical straights (sl: left, sr: right)

    Args:
        gap: gap between for coupler.
        gap_top: optional gap between top waveguides. Defaults to gap.
        radius: for the bend and coupler.
        length_x: ring coupler length.
        length_y: vertical straight length.
        coupler: ring coupler spec.
        bend: bend spec.
        straight: straight spec.
        cross_section: cross_section spec.

    .. code::

           o2──────▲─────────o3
                   │gap_top
           xx──────▼─────────xxx
          xxx                   xxx
        xxx                       xxx
       xx                           xxx
       x                             xxx
      xx                              xx▲
      xx                              xx│length_y
      xx                              xx▼
      xx                             xx
       xx          length_x          x
        xx     ◄───────────────►    x
         xx                       xxx
           xx                   xxx
            xxx──────▲─────────xxx
                     │gap
             o1──────▼─────────o4

    """
    gap_top = gap_top or gap
    gap = gf.snap.snap_to_grid(gap, grid_factor=2)
    gap_top = gf.snap.snap_to_grid(gap_top, grid_factor=2)

    xs = gf.get_cross_section(cross_section)
    radius = radius or xs.radius
    cross_section = xs.copy(radius=radius)

    coupler_component_top = gf.get_component(
        coupler_ring,
        gap=gap_top,
        radius=radius,
        length_x=length_x,
        bend=bend,
        cross_section=cross_section,
    )
    coupler_component_bot = gf.get_component(
        coupler_ring,
        gap=gap,
        radius=radius,
        length_x=length_x,
        bend=bend,
        cross_section=cross_section,
    )
    straight_component = straight(length=length_y, cross_section=cross_section)

    c = Component()
    cb = c.add_ref(coupler_component_bot)
    ct = c.add_ref(coupler_component_top)

    sl = straight_component.ref()
    sr = straight_component.ref()

    if length_y > 0:
        c.add(sl)
        c.add(sr)

    sl.connect(port="o1", destination=cb.ports["o2"])
    ct.connect(port="o3", destination=sl.ports["o2"])
    sr.connect(port="o2", destination=ct.ports["o2"])
    c.add_port("o1", port=cb.ports["o1"])
    c.add_port("o2", port=cb.ports["o4"])
    c.add_port("o3", port=ct.ports["o4"])
    c.add_port("o4", port=ct.ports["o1"])
    return c


if __name__ == "__main__":
    c = ring_double(length_x=10, radius=15, length_y=5)
    c.get_netlist()
    c.show(show_subports=False)



"""PDK stores layers, cross_sections, cell functions ..."""

from __future__ import annotations

import importlib
import pathlib
import warnings
from collections.abc import Callable
from functools import cached_property, partial
from typing import Any, Literal

import numpy as np
import omegaconf
from omegaconf import DictConfig
from pydantic import BaseModel, ConfigDict, Field, field_validator, validate_call

from gdsfactory.config import CONF, logger
from gdsfactory.events import Event
from gdsfactory.read.from_yaml_template import cell_from_yaml_template
from gdsfactory.show import show
from gdsfactory.symbols import floorplan_with_block_letters
from gdsfactory.technology import LayerStack, LayerViews, klayout_tech
from gdsfactory.typings import (
    CellSpec,
    Component,
    ComponentFactory,
    ComponentSpec,
    ConnectivitySpec,
    CrossSection,
    CrossSectionOrFactory,
    CrossSectionSpec,
    Layer,
    LayerSpec,
    MaterialSpec,
    PathType,
    Transition,
)

component_settings = ["function", "component", "settings"]
cross_section_settings = ["function", "cross_section", "settings"]

constants = {
    "fiber_array_spacing": 127.0,
    "fiber_spacing": 50.0,
    "fiber_input_to_output_spacing": 200.0,
    "metal_spacing": 10.0,
    "pad_spacing": 100.0,
    "pad_size": (80, 80),
}

nm = 1e-3


def evanescent_coupler_sample() -> None:
    """Evanescent coupler example.

    Args:
      coupler_length: length of coupling (min: 0.0, max: 200.0, um).
    """
    pass


def extract_args_from_docstring(docstring: str) -> dict[str, Any] | None:
    """
    This function extracts settings from a function's docstring for uPDK format.

    Args:
        docstring: The function from which to extract YAML in the docstring.

    Returns:
        settings (dict): The extracted YAML data as a dictionary.
    """
    args_dict = {}

    docstring_lines = docstring.split("\n") if docstring else []
    for line in docstring_lines:
        line = line.strip()
        if not line:
            continue
        if line.startswith("Args:"):
            continue
        if len(line.split(":")) != 2:
            continue
        name, description = line.split(":")
        name = name.strip()
        description_parts = description.split("(")
        doc = description_parts[0].strip()
        try:
            min_max_unit = description_parts[1].strip(")").split(",")
            min_val = float(min_max_unit[0].split(":")[1].strip())
            max_val = float(min_max_unit[1].split(":")[1].strip())
            unit = min_max_unit[2].strip()
        except IndexError:
            min_val = max_val = 0
            unit = None

        args_dict[name] = {
            "doc": doc,
            "min": min_val,
            "max": max_val,
            "type": "float",
            "unit": unit,
            "value": (min_val + max_val) / 2,  # setting default value as the midpoint
        }

    return args_dict


class CellDecoratorSettings(BaseModel):
    """Default settings for cell decorator."""

    max_name_length: int = Field(
        default=CONF.max_name_length,
        description="Maximum length of the cell name.",
    )
    include_module: bool = Field(
        default=False,
        description="If true, will include the module in the autogenerated component name.",
    )
    naming_style: Literal["default", "updk"] = Field(
        default="default", description="Naming style for autogenerated component names."
    )
    raise_error_on_mutation: bool = Field(default=CONF.raise_error_on_mutation)


class GdsWriteSettings(BaseModel):
    """Settings to use when writing to GDS."""

    on_uncached_component: Literal["warn", "error", "ignore"] = "ignore"
    lib_name: str = Field(
        default="library",
        description="Name of the GDS library to write to. Default is 'library'.",
    )
    unit: float = Field(
        default=1e-6,
        description="The units of coordinates in the database. The default is 1e-6 (1 micron).",
    )
    precision: float = Field(
        default=1e-9,
        description="The maximum precision of points in the database. For example, a value of 1e-9 would mean that you have a 1nm grid",
    )
    on_duplicate_cell: str = Field(
        default="warn",
        description="""Action to take when a duplicate cell is encountered on gds write (usually problematic). Options are
                        "warn" (default): overwrite all duplicate cells with one of the duplicates (arbitrarily).
                        "error": throw a ValueError when attempting to write a gds with duplicate cells.
                        "overwrite": overwrite all duplicate cells with one of the duplicates, without warning.""",
    )
    flatten_offgrid_references: bool = Field(
        default=False,
        description="If true, will auto-correct (and flatten) cell references which are off-grid or rotated by non-manhattan angles.",
    )
    max_points: int = Field(
        default=4000,
        description="Maximum number of points to allow in a polygon before fracturing.",
    )


class OasisWriteSettings(BaseModel):
    compression_level: int = Field(
        default=6,
        description="Level of compression for cells (between 0 and 9). Setting to 0 will disable cell compression, 1 gives the best speed, and 9 gives the best compression.",
    )
    detect_rectangles: bool = Field(
        default=True, description="If true, stores rectangles in a compressed format."
    )
    detect_trapezoids: bool = Field(
        default=True, description="If true, stores trapezoids in a compressed format."
    )
    circle_tolerance: float = Field(
        default=0,
        description="Tolerance for detecting circles. If less or equal to 0, no detection is performed. Circles are stored in compressed format.",
    )
    validation: str | None = Field(
        default=None,
        description="Type of validation to include in the saved file ('crc32', 'checksum32', or None).",
    )
    standard_properties: bool = Field(
        default=False,
        description="If true, stores standard OASIS properties in the file.",
    )


class Pdk(BaseModel):
    """Store layers, cross_sections, cell functions, simulation_settings ...

    only one Pdk can be active at a given time.

    Parameters:
        name: PDK name.
        cross_sections: dict of cross_sections factories.
        cells: dict of parametric cells that return Components.
        models: dict of models names to functions.
        symbols: dict of symbols names to functions.
        default_symbol_factory:
        base_pdk: a pdk to copy from and extend.
        base_pdks: list of pdks to copy from and extend.
        default_decorator: decorate all cells, if not otherwise defined on the cell.
        layers: maps name to gdslayer/datatype.
            For example dict(si=(1, 0), sin=(34, 0)).
        layer_stack: maps name to layer numbers, thickness, zmin, sidewall_angle.
            if can also contain material properties
            (refractive index, nonlinear coefficient, sheet resistance ...).
        layer_views: includes layer name to color, opacity and pattern.
        layer_transitions: transitions between different cross_sections.
        sparameters_path: to store Sparameters simulations.
        modes_path: to store Sparameters simulations.
        interconnect_cml_path: path to interconnect CML (optional).
        constants: dict of constants for the PDK.
        materials_index: material spec names to material spec, which can be:
            string: material name.
            float: refractive index.
            float, float: refractive index real and imaginary part.
            function: function of wavelength.
        routing_strategies: functions enabled to route.
        gds_write_settings: to write GDSII files.
        oasis_settings: to write OASIS files.
        cell_decorator_settings: settings for cell_without_validator decorator function in gdsfactory.cell.
        bend_points_distance: default points distance for bends in um.
        connectivity: defines connectivity between layers through vias.

    """

    name: str
    cross_sections: dict[str, CrossSectionOrFactory] = Field(
        default_factory=dict, exclude=True
    )
    cells: dict[str, ComponentFactory] = Field(default_factory=dict, exclude=True)
    models: dict[str, Callable] = Field(default_factory=dict, exclude=True)
    symbols: dict[str, ComponentFactory] = Field(default_factory=dict)
    default_symbol_factory: Callable = Field(
        default=floorplan_with_block_letters, exclude=True
    )
    base_pdk: Pdk | None = None
    base_pdks: list[Pdk] = Field(default_factory=list)
    default_decorator: Callable[[Component], None] | None = Field(
        default=None, exclude=True
    )
    layers: dict[str, Layer] = Field(default_factory=dict)
    layer_stack: LayerStack | None = None
    layer_views: LayerViews | None = None
    layer_transitions: dict[Layer | tuple[Layer, Layer], ComponentSpec] = Field(
        default_factory=dict
    )
    sparameters_path: PathType | None = Field(
        default=None, description="This field is deprecated."
    )

    modes_path: PathType | None = Field(
        default=None, description="This field is deprecated."
    )
    interconnect_cml_path: PathType | None = Field(
        default=None, description="This field is deprecated."
    )
    constants: dict[str, Any] = constants
    materials_index: dict[str, MaterialSpec] = Field(default_factory=dict)
    routing_strategies: dict[str, Callable] | None = None
    gds_write_settings: GdsWriteSettings = GdsWriteSettings()
    oasis_settings: OasisWriteSettings = OasisWriteSettings()
    cell_decorator_settings: CellDecoratorSettings = CellDecoratorSettings()
    bend_points_distance: float = 20 * nm
    connectivity: list[ConnectivitySpec] | None = None
    enforce_width_mismatch_layers: list[Layer] = Field(default_factory=list)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        ignore_extra=True,
        extra="forbid",
    )

    def __init__(self, **data):
        if "sparameters_path" in data:
            warnings.warn(
                "The 'pdk.sparameters_path' is deprecated. Use gf.config.PATH instead",
            )
        if "modes_path" in data:
            warnings.warn(
                "The 'pdk.modes_path' is deprecated. Use gf.config.PATH instead",
            )
        super().__init__(**data)

    @property
    def grid_size(self):
        """The minimum unit resolvable on the layout grid, relative to the unit."""
        return self.gds_write_settings.precision / self.gds_write_settings.unit

    @grid_size.setter
    def grid_size(self, value) -> None:
        self.gds_write_settings.precision = value * self.gds_write_settings.unit

    @field_validator("sparameters_path")
    def is_pathlib_path(cls, path):
        return pathlib.Path(path)

    def validate_layers(self, layers_required: list[Layer] | None = None):
        """Raises ValueError if layers_required are not in Pdk."""
        if layers_required is None:
            layers_required = []
        for layer in layers_required:
            if layer not in self.layers:
                raise ValueError(
                    f"{layer!r} not in Pdk.layers {list(self.layers.keys())}"
                )

    def activate(self) -> None:
        """Set current pdk to the active pdk (if not already active)."""
        global _ACTIVE_PDK
        if self is _ACTIVE_PDK:
            return None

        from gdsfactory.cell import clear_cache

        clear_cache()

        base_pdks = self.base_pdks

        if self.base_pdk:
            warnings.warn("base_pdk is deprecated. Use base_pdks instead")
            base_pdks += [self.base_pdk]

        for pdk in base_pdks:
            cross_sections = pdk.cross_sections
            cross_sections.update(self.cross_sections)
            cells = pdk.cells
            self.cross_sections = cross_sections
            cells.update(self.cells)
            self.cells.update(cells)

            layers = pdk.layers
            layers.update(self.layers)
            self.layers.update(layers)

            if not self.default_decorator:
                self.default_decorator = pdk.default_decorator

        layers_required = []
        self.validate_layers(layers_required)
        _set_active_pdk(self)

    def add_base_pdk(self):
        """Update pdk with self.base_pdk."""

    def register_cells(self, **kwargs) -> None:
        """Register cell factories."""
        for name, cell in kwargs.items():
            if not callable(cell):
                raise ValueError(
                    f"{cell} is not callable, make sure you register "
                    "cells functions that return a Component"
                )
            if name in self.cells:
                warnings.warn(f"Overwriting cell {name!r}")

            self.cells[name] = cell
            on_cell_registered.fire(name=name, cell=cell, pdk=self)

    def register_cross_sections(self, **kwargs) -> None:
        """Register cross_sections factories."""
        for name, cross_section in kwargs.items():
            if not callable(cross_section):
                raise ValueError(
                    f"{cross_section} is not callable, make sure you register "
                    "cross_section functions that return a CrossSection"
                )
            if name in self.cross_sections:
                warnings.warn(f"Overwriting cross_section {name!r}")
            self.cross_sections[name] = cross_section
            on_cross_section_registered.fire(
                name=name, cross_section=cross_section, pdk=self
            )

    def register_cells_yaml(
        self,
        dirpath: PathType | None = None,
        update: bool = False,
        **kwargs,
    ) -> None:
        """Load *.pic.yml YAML files and register them as cells.

        Args:
            dirpath: directory to recursive search for YAML cells.
            update: does not raise ValueError if cell already registered.

        Keyword Args:
            cell_name: cell function. To update cells dict.

        """

        message = "Updated" if update else "Registered"

        if dirpath:
            dirpath = pathlib.Path(dirpath)

            if not dirpath.is_dir():
                raise ValueError(f"{dirpath!r} needs to be a directory.")

            for filepath in dirpath.glob("**/*.pic.yml"):
                name = filepath.stem.split(".")[0]
                if not update and name in self.cells:
                    raise ValueError(
                        f"ERROR: Cell name {name!r} from {filepath} already registered."
                    )
                self.cells[name] = cell_from_yaml_template(filepath, name=name)
                on_yaml_cell_registered.fire(name=name, cell=self.cells[name], pdk=self)
                logger.info(f"{message} cell {name!r}")

        for k, v in kwargs.items():
            if not update and k in self.cells:
                raise ValueError(f"ERROR: Cell name {k!r} already registered.")
            self.cells[k] = v
            logger.info(f"{message} cell {k!r}")

    def remove_cell(self, name: str):
        """Removes cell from a PDK."""
        if name not in self.cells:
            raise ValueError(f"{name!r} not in {list(self.cells.keys())}")
        self.cells.pop(name)
        logger.info(f"Removed cell {name!r}")

    def get_cell(self, cell: CellSpec, **kwargs) -> ComponentFactory:
        """Returns ComponentFactory from a cell spec."""
        cells = set(self.cells.keys())

        if callable(cell):
            return cell
        elif isinstance(cell, str):
            if cell not in cells:
                cells = list(self.cells.keys())
                raise ValueError(
                    f"{cell!r} from PDK {self.name!r} not in cells: {cells} "
                )
            return self.cells[cell]
        elif isinstance(cell, dict | DictConfig):
            for key in cell.keys():
                if key not in component_settings:
                    raise ValueError(
                        f"Invalid setting {key!r} not in {component_settings}"
                    )
            settings = dict(cell.get("settings", {}))
            settings.update(**kwargs)

            cell_name = cell.get("function")
            if not isinstance(cell_name, str) or cell_name not in cells:
                cells = list(self.cells.keys())
                raise ValueError(
                    f"{cell_name!r} from PDK {self.name!r} not in cells: {cells} "
                )
            cell = self.cells[cell_name]
            return partial(cell, **settings)
        else:
            raise ValueError(
                "get_cell expects a CellSpec (ComponentFactory, string or dict),"
                f"got {type(cell)}"
            )

    def get_component(self, component: ComponentSpec, **kwargs) -> Component:
        """Returns component from a component spec."""
        return self._get_component(component=component, cells=self.cells, **kwargs)

    def get_symbol(self, component: ComponentSpec, **kwargs) -> Component:
        """Returns a component's symbol from a component spec."""
        # this is a pretty rough first implementation
        try:
            self._get_component(component=component, cells=self.symbols, **kwargs)
        except ValueError:
            component = self.get_component(component, **kwargs)
            return self.default_symbol_factory(component)

    def _get_component(
        self,
        component: ComponentSpec,
        cells: dict[str, Callable],
        validate: bool = False,
        **kwargs,
    ) -> Component:
        """Returns component from a component spec.

        Args:
            component: ComponentSpec (Component, ComponentFactory, string or dict).
            cells: dict of cells to search for component.
            validate: validate component with pydantic.
            kwargs: settings to override component settings.

        """
        cells = set(cells.keys())

        if isinstance(component, Component):
            if kwargs:
                warnings.warn(
                    f"Cannot apply settings {kwargs} to already created component {component.name!r}."
                    "Try passing a function instead of a Component."
                )
            return component
        elif callable(component):
            if validate:
                return validate_call(component)(**kwargs)
            else:
                return component(**kwargs)
        elif isinstance(component, str):
            if component not in cells:
                raise ValueError(
                    f"{component!r} not in PDK {self.name!r} cells: {cells} "
                )
            return self.cells[component](**kwargs)
        elif isinstance(component, dict | DictConfig):
            for key in component.keys():
                if key not in component_settings:
                    raise ValueError(
                        f"Invalid setting {key!r} not in {component_settings}"
                    )
            settings = dict(component.get("settings", {}))
            settings.update(**kwargs)

            cell_name = component.get("component", None)
            cell_name = cell_name or component.get("function")
            cell_name = cell_name.split(".")[-1]

            if not isinstance(cell_name, str) or cell_name not in cells:
                raise ValueError(
                    f"{cell_name!r} from PDK {self.name!r} not in cells: {cells} "
                )
            cell = self.cells[cell_name]
            if validate:
                return validate_call(cell)(**settings)
            else:
                return cell(**settings)
        else:
            raise ValueError(
                "get_component expects a ComponentSpec (Component, ComponentFactory, "
                f"string or dict), got {type(component)}"
            )

    def get_cross_section(
        self, cross_section: CrossSectionSpec, **kwargs
    ) -> CrossSection | Transition:
        """Returns cross_section from a cross_section spec."""
        if isinstance(cross_section, Transition):
            return cross_section
        elif isinstance(cross_section, CrossSection):
            return cross_section.copy(**kwargs)
        elif callable(cross_section):
            return cross_section(**kwargs)
        elif isinstance(cross_section, str):
            if cross_section not in self.cross_sections:
                cross_sections = list(self.cross_sections.keys())
                raise ValueError(f"{cross_section!r} not in {cross_sections}")
            xs = self.cross_sections[cross_section]
            return xs(**kwargs) if callable(xs) else xs.copy(**kwargs)
        elif isinstance(cross_section, dict | DictConfig):
            xs_name = cross_section.get("cross_section", None)
            settings = cross_section.get("settings", {})
            if kwargs:
                settings = settings | kwargs
            xs = self.get_cross_section(xs_name, **settings)
            return xs
        else:
            raise ValueError(
                "get_cross_section expects a CrossSectionSpec (CrossSection, "
                f"CrossSectionFactory, Transition, string or dict), got {type(cross_section)}"
            )

    def get_layer(self, layer: LayerSpec) -> Layer:
        """Returns layer from a layer spec."""
        if isinstance(layer, tuple | list):
            if len(layer) != 2:
                raise ValueError(f"{layer!r} needs two integer numbers.")
            return layer
        elif isinstance(layer, int):
            raise ValueError(
                f"A gds layer requires a tuple of two integers and got only one integer `{layer}`"
            )
        elif isinstance(layer, str):
            if layer not in self.layers:
                raise ValueError(f"{layer!r} not in {self.layers.keys()}")
            return self.layers[layer]
        elif layer is np.nan:
            return np.nan
        elif layer is None:
            return
        else:
            raise ValueError(
                f"{layer!r} needs to be a LayerSpec (string, int or Layer)"
            )

    def get_layer_views(self) -> LayerViews:
        if self.layer_views is None:
            raise ValueError(f"layer_views for Pdk {self.name!r} is None")
        return self.layer_views

    def get_layer_stack(self) -> LayerStack:
        if self.layer_stack is None:
            raise ValueError(f"layer_stack for Pdk {self.name!r} is None")
        return self.layer_stack

    def get_constant(self, key: str) -> Any:
        if not isinstance(key, str):
            return key
        if key not in self.constants:
            constants = list(self.constants.keys())
            raise ValueError(f"{key!r} not in {constants}")
        return self.constants[key]

    def get_material_index(self, key: str, *args, **kwargs) -> float:
        warnings.warn("get_material_index is deprecated")
        if key not in self.materials_index:
            material_names = list(self.materials_index.keys())
            raise ValueError(f"{key!r} not in {material_names}")
        material = self.materials_index[key]
        return material(*args, **kwargs) if callable(material) else material

    def to_updk(self) -> str:
        """Export to uPDK YAML definition."""
        from gdsfactory.components.bbox import bbox_to_points

        d = {}
        blocks = {cell_name: cell() for cell_name, cell in self.cells.items()}
        blocks = {
            name: dict(
                bbox=bbox_to_points(c.bbox),
                doc=c.__doc__.split("\n")[0]
                if c and hasattr(c, "__doc__") and c.__doc__ is not None
                else "",
                settings=extract_args_from_docstring(c.__doc__),
                parameters={
                    sname: {
                        "value": svalue,
                        "type": str(svalue.__class__.__name__),
                        "doc": extract_args_from_docstring(c.__doc__)
                        .get(sname, {})
                        .get("doc", None),
                        "min": extract_args_from_docstring(c.__doc__)
                        .get(sname, {})
                        .get("min", 0),
                        "max": extract_args_from_docstring(c.__doc__)
                        .get(sname, {})
                        .get("max", 0),
                        "unit": extract_args_from_docstring(c.__doc__)
                        .get(sname, {})
                        .get("unit", None),
                    }
                    for sname, svalue in dict(c.settings).items()
                    if isinstance(svalue, str | float | int)
                },
                pins={
                    port_name: {
                        "width": port.width,
                        "xsection": port.cross_section.name
                        if port.cross_section
                        else None,
                        "xya": [
                            float(port.center[0]),
                            float(port.center[1]),
                            float(port.orientation),
                        ],
                        "alias": port.info.get("alias"),
                        "doc": port.info.get("doc"),
                    }
                    for port_name, port in c.ports.items()
                },
            )
            for name, c in blocks.items()
        }
        xsections = {
            xs_name: self.get_cross_section(xs_name)
            for xs_name in self.cross_sections.keys()
        }
        xsections = {
            xs_name: dict(width=xsection.width)
            for xs_name, xsection in xsections.items()
        }

        header = dict(description=self.name)

        d["blocks"] = blocks
        d["xsections"] = xsections
        d["header"] = header
        return omegaconf.OmegaConf.to_yaml(d)

    # _on_cell_registered = Event()
    # _on_container_registered: Event = Event()
    # _on_yaml_cell_registered: Event = Event()
    # _on_cross_section_registered: Event = Event()
    #
    # @property
    # def on_cell_registered(self) -> Event:
    #     return self._on_cell_registered
    #
    # @property
    # def on_container_registered(self) -> Event:
    #     return self._on_container_registered
    #
    # @property
    # def on_yaml_cell_registered(self) -> Event:
    #     return self._on_yaml_cell_registered
    #
    # @property
    # def on_cross_section_registered(self) -> Event:
    #     return self._on_cross_section_registered

    def get_cross_section_name(self, cross_section: CrossSection) -> str:
        xs_name = next(
            (
                key
                for key, value in self.cross_sections.items()
                if value == cross_section
            ),
            None,
        )
        return xs_name or cross_section.name

    @cached_property
    def klayout_technology(self) -> klayout_tech.KLayoutTechnology:
        """Returns a KLayoutTechnology from the PDK.

        Raises:
            UserWarning if required properties for generating a KLayoutTechnology are not defined.
        """
        try:
            return klayout_tech.KLayoutTechnology(
                name=self.name,
                layer_views=self.layer_views,
                connectivity=self.connectivity,
                layer_map=self.layers,
                layer_stack=self.layer_stack,
            )
        except AttributeError as e:
            raise UserWarning(
                "Required properties for generating a KLayoutTechnology are not defined. "
                "Check the error for missing property"
            ) from e

    def __enter__(self) -> Pdk:
        """Called when entering the Pdk context manager."""
        self.activate()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Called when exiting the Pdk context manager."""
        global _ACTIVE_PDK
        _ACTIVE_PDK = None


_ACTIVE_PDK = None


def get_active_pdk(name: str | None = None) -> Pdk:
    """Returns active PDK.
    By default it will return the PDK defined in the name or config file.
    Otherwise it will return the generic PDK.
    """
    global _ACTIVE_PDK

    if _ACTIVE_PDK is None:
        if name is not None or CONF.pdk:
            pdk_module = importlib.import_module(name or CONF.pdk)
            pdk_module.PDK.activate()

        else:
            logger.debug("No active PDK. Activating generic PDK.\n")
            from gdsfactory.generic_tech import get_generic_pdk

            PDK = get_generic_pdk()
            PDK.activate()
            _ACTIVE_PDK = PDK
    return _ACTIVE_PDK


def get_material_index(material: MaterialSpec, *args, **kwargs) -> Component:
    return get_active_pdk().get_material_index(material, *args, **kwargs)


def get_component(component: ComponentSpec, **kwargs) -> Component:
    return get_active_pdk().get_component(component, **kwargs)


def get_cell(cell: CellSpec, **kwargs) -> ComponentFactory:
    return get_active_pdk().get_cell(cell, **kwargs)


def get_cross_section(
    cross_section: CrossSectionSpec, **kwargs
) -> CrossSection | Transition:
    return get_active_pdk().get_cross_section(cross_section, **kwargs)


def get_cross_section_name(cross_section: CrossSectionSpec) -> str:
    """Returns the name of a cross section. If not found in active PDK, raises a ValueError."""
    pdk = get_active_pdk()

    if isinstance(cross_section, str) and cross_section in pdk.cross_sections:
        return cross_section
    else:
        cross_section = get_cross_section(cross_section)
        for k, v in pdk.cross_sections.items():
            if cross_section == v:
                return k
        raise ValueError(f"Invalid cross section: {cross_section}")


def get_layer(layer: LayerSpec) -> Layer:
    return get_active_pdk().get_layer(layer)


def get_layer_views() -> LayerViews:
    return get_active_pdk().get_layer_views()


def get_layer_stack() -> LayerStack:
    return get_active_pdk().get_layer_stack()


def get_grid_size() -> float:
    return get_active_pdk().grid_size


def get_constant(constant_name: Any) -> Any:
    """If constant_name is a string returns a the value from the dict."""
    return (
        get_active_pdk().get_constant(constant_name)
        if isinstance(constant_name, str)
        else constant_name
    )


def get_sparameters_path() -> pathlib.Path:
    warnings.warn(
        "get_sparameters_path() is deprecated. gf.config.PATH.sparameters instead",
    )
    PDK = get_active_pdk()
    if PDK.sparameters_path is None:
        raise ValueError(f"{_ACTIVE_PDK.name!r} has no sparameters_path")
    return PDK.sparameters_path


def get_modes_path() -> pathlib.Path | None:
    warnings.warn(
        "get_modes_path() is deprecated. gf.config.PATH.modes instead",
    )
    PDK = get_active_pdk()
    return PDK.modes_path


def get_interconnect_cml_path() -> pathlib.Path:
    warnings.warn(
        "get_interconnect_cml_path() is deprecated. gf.config.PATH.interconnect_cml instead",
    )
    PDK = get_active_pdk()
    if PDK.interconnect_cml_path is None:
        raise ValueError(f"{_ACTIVE_PDK.name!r} has no interconnect_cml_path")
    return PDK.interconnect_cml_path


def _set_active_pdk(pdk: Pdk) -> None:
    global _ACTIVE_PDK
    old_pdk = _ACTIVE_PDK
    _ACTIVE_PDK = pdk
    on_pdk_activated.fire(old_pdk=old_pdk, new_pdk=pdk)


def get_routing_strategies() -> dict[str, Callable]:
    """Gets a dictionary of named routing functions available to the PDK, if defined, or gdsfactory defaults otherwise."""
    from gdsfactory.routing.factories import (
        routing_strategy as default_routing_strategies,
    )

    routing_strategies = get_active_pdk().routing_strategies
    if routing_strategies is None:
        routing_strategies = default_routing_strategies
    return routing_strategies


on_pdk_activated: Event = Event()
on_cell_registered: Event = Event()
on_container_registered: Event = Event()
on_yaml_cell_registered: Event = Event()
on_yaml_cell_modified: Event = Event()
on_cross_section_registered: Event = Event()

on_container_registered.add_handler(on_cell_registered.fire)
on_yaml_cell_registered.add_handler(on_cell_registered.fire)
on_yaml_cell_modified.add_handler(show)


if __name__ == "__main__":
    c = get_component("add_fiber_array")

    # from gdsfactory.read.from_updk import from_updk
    # from gdsfactory.samples.pdk.fab_c import pdk

    # yaml_pdk_decription = pdk.to_updk()
    # gdsfactory_script = from_updk(yaml_pdk_decription)
    # print(gdsfactory_script)
    # print(yaml_pdk_decription)

    # from gdsfactory.components import cells
    # from gdsfactory.cross_section import cross_sections

    # pdk = Pdk(
    #     name="demo",
    #     cells=cells,
    #     cross_sections=cross_sections,
    # layers=dict(DEVREC=(3, 0), PORTE=(3, 5)),
    # sparameters_path="/home",
    # )
    # print(pdk.json())
    # print(pdk.to_updk())



from __future__ import annotations

import gdsfactory as gf


@gf.cell
def mzi_phase_shifter(length_x: float = 100.0, **kwargs) -> gf.Component:
    """Returns MZI with TE grating couplers.

    Args:
        length_x: length of the phase shifter.
        kwargs: extra arguments passed to the MZI component.
    """
    test_info_mzi_heaters = dict(
        doe="mzis_heaters",
        analysis="mzi_heater",
        measurement="optical_loopback4_heater_sweep",
    )
    c = gf.components.mzi2x2_2x2_phase_shifter(length_x=length_x, **kwargs)
    c = gf.components.add_fiber_array_optical_south_electrical_north(
        c,
        electrical_port_names=["top_l_e2", "top_r_e2"],
    )
    c.info.update(test_info_mzi_heaters)
    return c


@gf.cell
def ring_single_heater(length_x: float = 10.0, **kwargs) -> gf.Component:
    """Returns ring with TE grating couplers.

    Args:
        length_x: length of the ring.
        kwargs: extra arguments passed to the ring component.
    """
    test_info_ring_heaters = dict(
        doe="ring_heaters",
        analysis="ring_heater",
        measurement="optical_loopback2_heater_sweep",
    )
    c = gf.components.ring_single_heater(length_x=length_x, **kwargs)
    c = gf.components.add_fiber_array_optical_south_electrical_north(
        c,
        electrical_port_names=["l_e2", "r_e2"],
    )
    c.info.update(test_info_ring_heaters)
    return c


@gf.cell
def spiral_sc(length: float = 20e3, **kwargs) -> gf.Component:
    """Returns spiral with TE grating couplers.

    Args:
        length: length of the spiral.
        kwargs: extra arguments passed to the spiral component.
    """
    info = dict(
        doe="spirals_sc",
        measurement="optical_loopback4",
        analysis="optical_loopback4_spirals",
    )
    c = gf.components.spiral_inner_io_fiber_array(length=length, **kwargs)
    c.info.update(info)
    return c


def sample_reticle(grid: bool = True, **kwargs) -> gf.Component:
    """Returns MZI with TE grating couplers.

    Args:
        grid: if True returns components on a regular grid. False packs them as close as possible.
        kwargs: passed to pack or grid.
    """

    mzis_te = [mzi_phase_shifter(length_x=length) for length in [100, 200, 300]]
    rings_te = [ring_single_heater(length_x=length_x) for length_x in [10, 20, 30]]
    spirals_te = [spiral_sc(length=length) for length in [20e3, 40e3, 60e3]]

    components = mzis_te + rings_te + spirals_te

    if grid:
        return gf.grid(components, name_ports_with_component_name=True, **kwargs)
    c = gf.pack(components, **kwargs)
    if len(c) > 1:
        raise ValueError(f"failed to pack into single group. Made {len(c)} groups.")
    return c[0]


if __name__ == "__main__":
    c = sample_reticle(grid=False)
    # c = mzi_phase_shifter()
    # c = spiral_sc()
    c.show(show_ports=True)
    # c.pprint_ports()



"""Straight waveguide."""

from __future__ import annotations

import warnings

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.cross_section import CrossSectionSpec


@gf.cell
def straight(
    length: float = 10.0,
    npoints: int = 2,
    cross_section: CrossSectionSpec = "xs_sc",
    **kwargs,
) -> Component:
    """Returns a Straight waveguide.

    Args:
        length: straight length (um).
        npoints: number of points.
        cross_section: specification (CrossSection, string or dict).
        kwargs: additional cross_section arguments.

    .. code::

        o1 -------------- o2
                length
    """
    c = Component()

    x = gf.get_cross_section(cross_section, **kwargs)
    p = gf.path.straight(length=length, npoints=npoints)
    path = p.extrude(x)
    ref = c << path
    c.add_ports(ref.ports)

    c.info["length"] = float(length)
    if len(x.sections) == 0:
        c.info["width"] = x.width
    else:
        c.info["width"] = x.sections[0].width

    c.add_route_info(cross_section=x, length=length)
    c.absorb(ref)
    return c


@gf.cell
def straight_array(
    n: int = 4,
    spacing: float = 4.0,
    length: float = 10.0,
    cross_section: CrossSectionSpec = "xs_sc",
) -> Component:
    """Array of straights connected with grating couplers.

    useful to align the 4 corners of the chip

    Args:
        n: number of straights.
        spacing: edge to edge straight spacing.
        length: straight length (um).
        cross_section: specification (CrossSection, string or dict).
    """
    warnings.warn("Use gf.components.array(straight) instead", DeprecationWarning)

    c = Component()
    wg = straight(cross_section=cross_section, length=length)

    for i in range(n):
        wref = c.add_ref(wg)
        wref.y += i * (spacing + wg.info["width"])
        c.add_ports(wref.ports, prefix=str(i))

    c.auto_rename_ports()
    return c


if __name__ == "__main__":
    import gdsfactory as gf

    # xs = gf.cross_section.strip(bbox_layers=[(111, 0)], bbox_offsets=[3])
    # c = straight(cross_section=xs, info=dict(simulation="eme"))
    c = straight(cross_section="xs_sc_pins")
    # print(c.info["simulation"])
    # c = gf.Component()
    # ref = c << straight(width=3e-3, length=3e-3)
    # ref.xmin = 0
    # ref.ymin = 0
    # ref.center = (0, 0)

    c.show()



"""Converts CSV of test site labels into a CSV test manifest."""

import json
import pathlib
import warnings
from functools import partial

import pandas as pd

import gdsfactory as gf

marker = partial(gf.components.rectangle, layer="TEXT", centered=True, size=(10, 10))


def write_test_manifest(
    csvpath: str | pathlib.Path,
    gdspath: str | pathlib.Path | None = None,
    marker_optical=marker,
    marker_electrical=marker,
) -> pd.DataFrame:
    """Converts CSV of test site labels into a CSV test manifest.

    Args:
        csvpath: path to CSV file with test site labels.
        gdspath: path to GDS file with test site labels.
        marker_optical: marker to use for test site labels.
        marker_electrical: marker to use for test site labels.
    """

    warnings.warn(
        "This function is deprecated. Use gf.labels.get_test_manifest instead",
        DeprecationWarning,
    )

    df_in = pd.read_csv(csvpath)

    # Initialize an empty list to collect the rows
    rows = []
    columns = [
        "cell",
        "device",
        "port_type",
        "port_names",
        "x",
        "y",
        "measurement",
        "measurement_settings",
        "doe",
        "analysis",
        "analysis_settings",
        "cell_settings",
    ]
    if gdspath:
        c = gf.Component()
        _ = c << gf.import_gds(gdspath)
        marker_optical = marker_optical()
        marker_electrical = marker_electrical()

    for _, label in df_in.iterrows():
        x, y, _orientation = label.x, label.y, label.rotation
        text = label.text
        d = json.loads(text)

        if gdspath:
            if d["port_type"] == "optical":
                ref = c << marker_optical
                ref.x = x
                ref.y = y
            elif d["port_type"] == "electrical":
                ref = c << marker_electrical
                ref.x = x
                ref.y = y

        row = [
            d["name"],
            d["name"] + f"_{int(x)}_{int(y)}",
            d["port_type"],
            d["port_names"],
            x,
            y,
            d["measurement"],
            d["measurement_settings"],
            d["doe"],
            d["analysis"],
            d["analysis_settings"],
            d["cell_settings"],
        ]
        rows.append(row)

    if gdspath:
        c.show()

    return pd.DataFrame(
        rows,
        columns=columns,
    )


if __name__ == "__main__":
    test_info_mzi_heaters = dict(
        doe="mzis_heaters",
        analysis="mzi_heater_phase_shifter_length",
        measurement="optical_loopback4_heater_sweep",
    )
    test_info_ring_heaters = dict(
        doe="ring_heaters_coupling_length",
        analysis="ring_heater",
        measurement="optical_loopback2_heater_sweep",
    )

    mzis = [
        gf.components.mzi2x2_2x2_phase_shifter(length_x=lengths)
        for lengths in [100, 200, 300]
    ]

    rings = [
        gf.components.ring_single_heater(length_x=length_x) for length_x in [10, 20, 30]
    ]

    mzis_te = [
        gf.components.add_fiber_array_optical_south_electrical_north(
            mzi,
            electrical_port_names=["top_l_e2", "top_r_e2"],
            info=test_info_mzi_heaters,
            decorator=gf.labels.add_label_json,
        )
        for mzi in mzis
    ]
    rings_te = [
        gf.components.add_fiber_array_optical_south_electrical_north(
            ring,
            electrical_port_names=["l_e2", "r_e2"],
            info=test_info_ring_heaters,
            decorator=gf.labels.add_label_json,
        )
        for ring in rings
    ]
    c = gf.pack(mzis_te + rings_te)[0]
    c.show(show_ports=False)
    gdspath = c.write_gds("sample_reticle.gds")
    csvpath = gf.labels.write_labels.write_labels_gdstk(
        gdspath, prefixes=("{",), layer_label="TEXT"
    )
    df = write_test_manifest(csvpath, gdspath=gdspath)
    df.to_csv("test_manifest.csv")
    print(df)



import gdsfactory as gf
import gdsfactory.typings as gt

if __name__ == "__main__":
    import matplotlib.pyplot as plt

    s = gt.Schematic()
    s.add_instance("mzi1", gt.Instance(component=gf.c.mzi(delta_length=10)))
    s.add_instance("mzi2", gt.Instance(component=gf.c.mzi(delta_length=100)))
    s.add_instance("mzi3", gt.Instance(component=gf.c.mzi(delta_length=200)))
    s.add_placement("mzi1", gt.Placement(x=000))
    s.add_placement("mzi2", gt.Placement(x=100))
    s.add_placement("mzi3", gt.Placement(x=200))
    s.add_net(gt.Net(ip1="mzi1,o2", ip2="mzi2,o2"))
    s.add_net(gt.Net(ip1="mzi2,o2", ip2="mzi3,o1"))
    g = s.plot_netlist()
    plt.show()



from pathlib import Path

import kfactory as kf

import gdsfactory as gf


def import_gds(gdspath: str | Path) -> kf.KCell:
    """Reads a GDS file and returns a kfactory component.

    Args:
        gdspath: path to GDS file.
    """
    read_options = kf.kcell.load_layout_options()
    read_options.cell_conflict_resolution = (
        kf.kdb.LoadLayoutOptions.CellConflictResolution.RenameCell
    )
    top_cells = set(kf.kcl.top_cells())
    kf.kcl.read(gdspath, read_options, test_merge=True)
    new_top_cells = set(kf.kcl.top_cells()) - top_cells
    if len(new_top_cells) != 1:
        raise ValueError(f"Expected 1 new top cell, got {len(new_top_cells)}")
    return kf.kcl[new_top_cells.pop().name]


def add_ports(c: kf.KCell, gf_component: gf.Component):
    """Adds ports to a kfactory component.

    Args:
        c: kfactory component.
        gf_component: gdsfactory component.
    """
    for port in gf_component.ports.values():
        center = port.center
        orientation = port.orientation or 0
        layer = gf.get_layer(port.layer)
        layer = kf.kcl.layer(*layer)
        trans = kf.kdb.DCplxTrans(1, float(orientation), False, center[0], center[1])
        width = port.width
        c.create_port(
            name=port.name,
            dwidth=round(width * 1e3) / 1e3,
            layer=layer,
            port_type=port.port_type,
            dcplx_trans=trans,
        )


def to_kfactory(component: gf.Component, recursive: bool = True) -> kf.KCell:
    """Converts a gdsfactory component to a kfactory component.

    Args:
        component: gdsfactory component.
        recursive: if True, recursively add ports and info to all subcomponents.

    """
    c0 = component.write_gds(with_metadata=True)
    c = import_gds(c0)

    for k, v in sorted(dict(component.info).items()):
        c.info[k] = v

    settings = {}
    for k, v in sorted(dict(component.settings).items()):
        if isinstance(v, str | int | float | bool):
            settings[k] = v

    c._settings = kf.kcell.KCellSettings(**settings)
    add_ports(c, component)

    if recursive:
        for child in component.references:
            component = child.parent
            cell_name = component.name
            kcell_child = kf.kcl[cell_name]
            add_ports(kcell_child, component)

            for k, v in sorted(dict(component.info).items()):
                kcell_child.info[k] = v

            settings = {}
            for k, v in sorted(dict(component.settings).items()):
                if isinstance(v, str | int | float | bool):
                    settings[k] = v
            kcell_child._settings = kf.kcell.KCellSettings(**settings)

    return c


if __name__ == "__main__":
    # c0 = gf.components.mzi()
    c0 = gf.pack([gf.c.mzi(), gf.c.straight()])[0]
    c = to_kfactory(c0)
    c.show()
    # c = gf.components.mzi()
    # c.write_gds(with_metadata=True)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.cross_section import CrossSectionSpec
from gdsfactory.typings import ComponentSpec


@gf.cell
def coupler_straight(
    length: float = 10.0,
    gap: float = 0.27,
    straight: ComponentSpec = straight_function,
    cross_section: CrossSectionSpec = "xs_sc",
    **kwargs,
) -> Component:
    """Coupler_straight with two parallel straights.

    Args:
        length: of straight.
        gap: between straights.
        straight: straight component (straight, bend_euler, bend_heater).
        cross_section: specification (CrossSection, string or dict).
        kwargs: cross_section settings.

    .. code::

        o2──────▲─────────o3
                │gap
        o1──────▼─────────o4
    """
    component = Component()

    straight_component = straight(length=length, cross_section=cross_section, **kwargs)

    top = component << straight_component
    bot = component << straight_component
    top.movey(straight_component.info["width"] + gap)

    component.add_port("o1", port=bot.ports["o1"])
    component.add_port("o2", port=top.ports["o1"])
    component.add_port("o3", port=bot.ports["o2"])
    component.add_port("o4", port=top.ports["o2"])
    component.auto_rename_ports()
    return component


if __name__ == "__main__":
    c = coupler_straight(length=2)
    c.show(show_ports=False)



from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.path import arc
from gdsfactory.snap import snap_to_grid
from gdsfactory.typings import CrossSectionSpec


@gf.cell
def bend_circular(
    radius: float | None = None,
    angle: float = 90.0,
    npoints: int | None = None,
    cross_section: CrossSectionSpec = "xs_sc",
    **kwargs,
) -> Component:
    """Returns a radial arc.

    Args:
        radius: in um. Defaults to cross_section_radius.
        angle: angle of arc (degrees).
        npoints: number of points.
        layer: layer to use. Defaults to cross_section.layer.
        width: width to use. Defaults to cross_section.width.
        cross_section: spec (CrossSection, string or dict).
        kwargs: additional cross_section arguments.

    .. code::

                  o2
                  |
                 /
                /
               /
       o1_____/
    """
    x = gf.get_cross_section(cross_section, **kwargs)
    radius = radius or x.radius

    p = arc(radius=radius, angle=angle, npoints=npoints)
    c = Component()
    path = p.extrude(x)
    ref = c << path
    c.add_ports(ref.ports)
    c.absorb(ref)

    c.info["length"] = float(snap_to_grid(p.length()))
    c.info["dy"] = snap_to_grid(float(abs(p.points[0][0] - p.points[-1][0])))
    c.info["radius"] = float(radius)
    x.validate_radius(radius)

    c.add_route_info(
        cross_section=x, length=c.info["length"], n_bend_90=abs(angle / 90.0)
    )
    return c


bend_circular180 = partial(bend_circular, angle=180)


if __name__ == "__main__":
    import gdsfactory as gf

    x1 = gf.cross_section.strip(width=1)
    x2 = gf.cross_section.strip(width=2)
    x = gf.cross_section.Transition(cross_section1=x1, cross_section2=x2)

    c = bend_circular(angle=180, cross_section=x, radius=10)
    c.show(show_ports=True)



from __future__ import annotations

from collections.abc import Callable

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.taper_cross_section import taper_cross_section
from gdsfactory.cross_section import strip
from gdsfactory.port import select_ports_optical
from gdsfactory.typings import ComponentFactory, ComponentSpec, CrossSectionSpec


@cell
def add_tapers(
    component: ComponentSpec,
    taper: ComponentFactory | Component = taper_cross_section,
    select_ports: Callable | None = select_ports_optical,
    taper_port_name1: str = "o1",
    taper_port_name2: str = "o2",
    cross_section1: CrossSectionSpec | None = None,
    cross_section2: CrossSectionSpec = strip,
    **kwargs,
) -> Component:
    """Returns new component with taper in all optical ports.

    Args:
        component: to add tapers.
        taper: taper spec.
        select_ports: function to select ports.
        taper_port_name1: name.
        taper_port_name2: name.
        cross_section1: optional start cross_section factory. Defaults to port.
        cross_section2: end cross_section factory (cross_section).

    Keyword Args:
        length: transition length.
        npoints: number of points.
        linear: shape of the transition, sine when False.
        kwargs: cross_section settings for section2.

    Note:
        If ``taper`` is a partial function and ``cross_section2`` is None, then
        ``cross_section2`` is inferred from the partial keywords.
    """
    c = gf.Component()
    component = gf.get_component(component)

    ports_to_taper = select_ports(component.ports) if select_ports else component.ports
    ports_to_taper_names = [p.name for p in ports_to_taper.values()]

    for port_name, port in component.ports.items():
        if port.name in ports_to_taper_names:
            if (
                isinstance(taper, gf.partial)
                and "cross_section2" in taper.keywords
                and cross_section2 is None
            ):
                _taper = gf.get_component(
                    taper,
                    cross_section2=gf.partial(
                        taper.keywords["cross_section2"], width=port.width
                    ),
                )
            elif isinstance(taper, Component):
                _taper = taper
            else:
                _taper = gf.get_component(
                    taper,
                    cross_section1=cross_section1 or port.cross_section,
                    cross_section2=cross_section2,
                    **kwargs,
                )
            taper_ref = c << _taper
            taper_ref.connect(taper_ref.ports[taper_port_name1].name, port)
            c.add_port(name=port_name, port=taper_ref.ports[taper_port_name2])
        else:
            c.add_port(name=port_name, port=port)
    c.add_ref(component)
    c.copy_child_info(component)
    return c


if __name__ == "__main__":
    c0 = gf.components.straight(width=2, cross_section=gf.cross_section.rib)
    xs_rc_tip = gf.cross_section.strip_rib_tip

    c1 = add_tapers(c0, cross_section2=xs_rc_tip, linear=True)
    c1.show()



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import LayerSpec, LayerSpecs


@gf.cell
def cross(
    length: float = 10.0,
    width: float = 3.0,
    layer: LayerSpec = "WG",
    layers: LayerSpecs | None = None,
    port_type: str | None = None,
) -> Component:
    """Returns a cross from two rectangles of length and width.

    Args:
        length: float Length of the cross from one end to the other.
        width: float Width of the arms of the cross.
        layer: layer for geometry.
        layers: Optional. List of layers for geometry.
        port_type: None, optical, electrical.
    """
    c = gf.Component()
    layers = layers or [layer]

    for layer in layers:
        layer = gf.get_layer(layer)
        R = gf.components.rectangle(size=(width, length), layer=layer)
        r1 = c.add_ref(R).rotate(90)
        r2 = c.add_ref(R)
        r1.center = (0, 0)
        r2.center = (0, 0)

        if port_type:
            c.add_port(
                1,
                width=width,
                layer=layer,
                orientation=0,
                center=(+length / 2, 0),
                port_type=port_type,
            )
            c.add_port(
                2,
                width=width,
                layer=layer,
                orientation=180,
                center=(-length / 2, 0),
                port_type=port_type,
            )
            c.add_port(
                3,
                width=width,
                layer=layer,
                orientation=90,
                center=(0, length / 2),
                port_type=port_type,
            )
            c.add_port(
                4,
                width=width,
                layer=layer,
                orientation=270,
                center=(0, -length / 2),
                port_type=port_type,
            )
        c.auto_rename_ports()
    return c


if __name__ == "__main__":
    layers = {(1, 0), (2, 0)}
    c = cross(layers=layers)
    c.show(show_ports=True)
    # c.pprint_ports()
    # cc = gf.routing.add_fiber_array(component=c)
    # cc.show(show_ports=True)



"""Design For Testing module includes test protocols."""

from __future__ import annotations

from gdsfactory.labels import siepic, write_labels
from gdsfactory.labels.add_label_yaml import add_label_json, add_label_yaml
from gdsfactory.labels.ehva import add_label_ehva
from gdsfactory.labels.get_test_manifest import get_test_manifest
from gdsfactory.labels.siepic import add_fiber_array_siepic

__all__ = [
    "add_fiber_array_siepic",
    "add_label_ehva",
    "add_label_yaml",
    "add_label_json",
    "siepic",
    "write_labels",
    "get_test_manifest",
]



"""
Some manhattan routes have disconnected waveguides

"""

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    # c1 = c << gf.components.straight_array(spacing=200)
    c1 = c << gf.components.straight_array(spacing=50)
    c2 = c << gf.components.straight_array(spacing=5)

    c2.movex(200)
    c1.y = 0
    c2.y = 0

    routes = gf.routing.get_bundle_path_length_match(
        c1.get_ports_list(orientation=0),
        c2.get_ports_list(orientation=180),
        end_straight_length=0,
        start_straight_length=0,
        separation=50,
        radius=10,
        # radius=3 # smaller radius works
        enforce_port_ordering=False,
    )

    for route in routes:
        c.add(route.references)

    c.show()



from __future__ import annotations

import numpy as np

import gdsfactory as gf
from gdsfactory.typings import Float2, LayerSpec, LayerSpecs


@gf.cell
def fiducial_squares(
    layer: LayerSpec = "WG",
    layers: LayerSpecs | None = None,
    size: Float2 = (5.0, 5.0),
    offset: float = 0.14,
) -> gf.Component:
    """Returns fiducials with two squares.

    Args:
        layer: layer for geometry.
        layers: optional list of layers to duplicate the geometry.
        size: in um.
        offset: between squares in um.
    """
    c = gf.Component()
    layers = layers or [layer]

    for layer in layers:
        layer = gf.get_layer(layer)
        rectangle = gf.components.rectangle(size=size, layer=layer)
        c << rectangle
        r2 = c << rectangle
        r2.move(-np.array(size) - np.array([offset, offset]))

    return c


if __name__ == "__main__":
    c = fiducial_squares()
    c.show()



from __future__ import annotations

from functools import partial

import numpy as np

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.circle import circle
from gdsfactory.components.compass import compass
from gdsfactory.typings import Ints, LayerSpec


@cell
def rectangle(
    size=(4.0, 2.0),
    layer: LayerSpec = "WG",
    centered: bool = False,
    port_type: str | None = "electrical",
    port_orientations: Ints | None = (180, 90, 0, -90),
    round_corners_east_west: bool = False,
    round_corners_north_south: bool = False,
) -> Component:
    """Returns a rectangle.

    Args:
        size: (tuple) Width and height of rectangle.
        layer: Specific layer to put polygon geometry on.
        centered: True sets center to (0, 0), False sets south-west to (0, 0).
        port_type: optical, electrical.
        port_orientations: list of port_orientations to add.
        round_corners_east_west: if True adds rounded east / west corners using circles at the edges.
        round_corners_north_south: if True adds rounded north / south corners using circles at the edges.
    """
    c = Component()
    xsize, ysize = size

    if round_corners_east_west:
        xsize -= ysize
    if round_corners_north_south:
        ysize -= xsize

    ref = c << compass(
        size=(xsize, ysize),
        layer=layer,
        port_type=port_type,
        port_orientations=port_orientations,
    )
    if not centered:
        ref.move((xsize / 2, ysize / 2))
    if port_type:
        c.add_ports(ref.ports)

    if round_corners_east_west:
        _circle = circle(radius=ysize / 2, layer=layer)
        cl = c << _circle
        cr = c << _circle
        cl.x = ref.xmin
        cr.x = ref.xmax
        cr.y = ref.y
        cl.y = ref.y

    if round_corners_north_south:
        _circle = circle(radius=xsize / 2, layer=layer)
        ct = c << _circle
        cb = c << _circle
        ct.x = ref.x
        cb.x = ref.x
        cb.y = ref.ymin
        ct.y = ref.ymax

    c.info["area"] = float(xsize * ysize)
    return c


fiber_size = 10.4
marker_te = partial(rectangle, size=[fiber_size, fiber_size], layer="TE", centered=True)
marker_tm = partial(rectangle, size=[fiber_size, fiber_size], layer="TM", centered=True)


@cell
def rectangles(
    size=(4.0, 2.0),
    offsets=(0, 1),
    layers=("WG", "SLAB150"),
    centered: bool = True,
    **kwargs,
) -> Component:
    """Returns overimposed rectangles.

    Args:
        size: (tuple) Width and height of rectangle.
        layers: Specific layer to put polygon geometry on.
        offsets: list of offsets.
        centered: True sets center to (0, 0), False sets south-west of first rectangle to (0, 0).

    Keyword Args:
        port_type: optical, electrical.
        port_orientations: list of port_orientations to add.

    .. code::

            ┌──────────────┐
            │              │
            │   ┌──────┐   │
            │   │      │   │
            │   │      ├───►
            │   │      │offset
            │   └──────┘   │
            │              │
            └──────────────┘

    """
    c = Component()
    size = np.array(size)

    ref0 = None
    kwargs.pop("layer", None)

    if len(offsets) != len(layers):
        raise ValueError(f"len(offsets) != len(layers) {len(offsets)} != {len(layers)}")
    for layer, offset in zip(layers, offsets):
        ref = c << rectangle(
            size=size + 2 * offset, layer=layer, centered=centered, **kwargs
        )
        if ref0:
            ref.center = ref0.center
        ref0 = ref

    return c


if __name__ == "__main__":
    layer = (1, 0)
    # c = rectangles(offsets=(0, 1), centered=False, layer=layer)
    c = rectangle(
        size=(20, 30), centered=False, layer=layer, round_corners_north_south=True
    )
    # c = rectangle(size=(3, 2), centered=True, layer=(2, 3))
    print(c.ports)
    print(c.name)
    c.show(show_ports=False)



"""You can define a function to add pins."""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.add_pins import add_pins_triangle


@gf.cell(post_process=[add_pins_triangle])
def straight(**kwargs) -> gf.Component:
    return gf.components.straight(**kwargs)


if __name__ == "__main__":
    c = straight()
    c.show(show_ports=False)



"""Cell decorator for functions that return a Component."""

from __future__ import annotations

import functools
import hashlib
import inspect
import warnings
from collections.abc import Callable, Sequence
from functools import partial
from typing import TypeVar, overload

from pydantic import validate_call

from gdsfactory.component import Component, name_counters
from gdsfactory.component_layout import CellSettings
from gdsfactory.config import CONF
from gdsfactory.name import clean_name, get_name_short
from gdsfactory.serialization import clean_value_name

CACHE: dict[str, Component] = {}
CACHE_IDS = set()

INFO_VERSION = 2

_F = TypeVar("_F", bound=Callable)


class CellReturnTypeError(ValueError):
    pass


def remove_from_cache(name: str | Component) -> None:
    """Removes Component name from CACHE and resets the name counter."""

    if not isinstance(name, str):
        name = name.name

    if name in CACHE:
        del CACHE[name]

    if name_counters[name] == 1:
        name_counters[name] = 0


def clear_cache() -> None:
    """Clears Component CACHE and reset the name counters."""

    CACHE.clear()
    CACHE_IDS.clear()
    name_counters.clear()


def print_cache() -> None:
    for k in CACHE:
        print(k)


@overload
def cell(
    func: None,
    /,
    *,
    autoname: bool = True,
    max_name_length: int | None = None,
    include_module: bool = False,
    with_hash: bool = False,
    ports_offgrid: str | None = None,
    ports_not_manhattan: str | None = None,
    flatten: bool = False,
    naming_style: str = "default",
    default_decorator: Callable[[Component], Component] | None = None,
    add_settings: bool = True,
    validate: bool = False,
    get_child_name: bool = False,
    post_process: Sequence[Callable] | None = None,
    info: dict[str, int | float | str] | None = None,
) -> partial: ...


@overload
def cell(
    func: Callable[..., Component],
    /,
    *,
    autoname: bool = True,
    max_name_length: int | None = None,
    include_module: bool = False,
    with_hash: bool = False,
    ports_offgrid: str | None = None,
    ports_not_manhattan: str | None = None,
    flatten: bool = False,
    naming_style: str = "default",
    default_decorator: Callable[[Component], Component] | None = None,
    add_settings: bool = True,
    validate: bool = False,
    get_child_name: bool = False,
    post_process: Sequence[Callable] | None = None,
    info: dict[str, int | float | str] | None = None,
) -> Callable[..., Component]: ...


def cell(
    func: Callable[..., Component] | None = None,
    /,
    *,
    autoname: bool = True,
    max_name_length: int | None = None,
    include_module: bool = False,
    with_hash: bool = False,
    ports_offgrid: str | None = None,
    ports_not_manhattan: str | None = None,
    flatten: bool = False,
    naming_style: str = "default",
    default_decorator: Callable[[Component], Component] | None = None,
    add_settings: bool = True,
    validate: bool = False,
    get_child_name: bool = False,
    post_process: Sequence[Callable] | None = None,
    info: dict[str, int | float | str] | None = None,
) -> Callable[..., Component] | partial:
    """Parametrized Decorator for Component functions.

    Args:
        func: function to decorate.
        autoname: True renames Component based on args and kwargs. True by default.
        max_name_length: truncates name beyond some characters with a hash. Defaults to CONF.max_name_length.
        include_module: True adds module name to the cell name.
        with_hash: True adds a hash to the cell name.
        ports_offgrid: "warn", "error" or "ignore". Checks if ports are on grid. Defaults to CONF.ports_offgrid.
        ports_not_manhattan: "warn", "error" or "ignore". Checks if ports are manhattan. Defaults to CONF.ports_non_manhattan.
        flatten: False by default. True flattens component hierarchy.
        naming_style: "default" or "updk". "default" is the default naming style.
        default_decorator: default decorator to apply to the component. None by default.
        add_settings: True by default. Adds settings to the component.
        validate: validate the function call. Does not work with annotations that have None | Callable.
        get_child_name: Use child name as component name prefix.
        post_process: list of post processing functions to apply to the component.
        info: dictionary with metadata to add to the component.

    Implements a cache so that if a component has already been build it returns the component from the cache directly.
    This avoids creating two exact Components that have the same name.
    Can autoname components based on the function name and arguments.

    A decorator is a function that runs over a function, so when you do.

    .. code::

        import gdsfactory as gf

        @gf.cell
        def mzi_with_bend():
            c = gf.Component()
            mzi = c << gf.components.mzi()
            bend = c << gf.components.bend_euler()
            return c

    it’s equivalent to

    .. code::

        def mzi_with_bend():
            c = gf.Component()
            mzi = c << gf.components.mzi()
            bend = c << gf.components.bend_euler(radius=radius)
            return c

        mzi_with_bend_decorated = gf.cell(mzi_with_bend)

    """
    if func is None:
        return partial(
            cell,
            autoname=autoname,
            max_name_length=max_name_length,
            include_module=include_module,
            with_hash=with_hash,
            ports_offgrid=ports_offgrid,
            ports_not_manhattan=ports_not_manhattan,
            flatten=flatten,
            naming_style=naming_style,
            default_decorator=default_decorator,
            add_settings=add_settings,
            validate=validate,
            get_child_name=get_child_name,
            post_process=post_process,
            info=info,
        )

    if default_decorator is not None:
        warnings.warn(
            "default_decorator is deprecated and will be removed soon. Use post_process instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Component:
        assert func is not None
        nonlocal ports_not_manhattan, ports_offgrid, max_name_length
        from gdsfactory.pdk import get_active_pdk

        active_pdk = get_active_pdk()

        name = kwargs.pop("name", None)
        prefix = kwargs.pop("prefix", None)
        metadata = info or {}  # noqa

        if name:
            warnings.warn(
                f"name is deprecated and will be removed soon. {func.__name__}",
                stacklevel=2,
            )
        if prefix:
            warnings.warn(
                f"prefix is deprecated and will be removed soon. {func.__name__}",
                stacklevel=2,
            )

        prefix = prefix or func.__name__
        sig = inspect.signature(func)
        args_as_kwargs = dict(zip(sig.parameters.keys(), args))
        args_as_kwargs.update(kwargs)

        if max_name_length is None:
            max_name_length = CONF.max_name_length
        if ports_offgrid is None:
            ports_offgrid = CONF.ports_offgrid
        if ports_not_manhattan is None:
            ports_not_manhattan = CONF.ports_not_manhattan

        default = {
            p.name: p.default
            for p in sig.parameters.values()
            if p.default != inspect._empty
        }

        changed = args_as_kwargs
        full = default.copy()
        full.update(**args_as_kwargs)
        default2 = default.copy()
        changed2 = changed.copy()

        # list of default args as strings
        default_args_list = [
            f"{key}={clean_value_name(default2[key])}" for key in sorted(default.keys())
        ]
        # list of explicitly passed args as strings
        passed_args_list = [
            f"{key}={clean_value_name(changed2[key])}" for key in sorted(changed.keys())
        ]

        if naming_style == "updk":
            full_args_list = [
                f"{key}={clean_value_name(full[key])}" for key in sorted(full.keys())
            ]
            named_args_string = ",".join(full_args_list)
            name = f"{prefix}:{named_args_string}" if named_args_string else prefix
            name = clean_name(name, allowed_characters=[":", ".", "="])

        elif naming_style == "default":
            changed_arg_set = set(passed_args_list).difference(default_args_list)
            changed_arg_list = sorted(changed_arg_set)
            named_args_string = "_".join(changed_arg_list)

            if include_module:
                named_args_string += f"_{func.__module__}"
            if changed_arg_list:
                named_args_string = (
                    hashlib.md5(named_args_string.encode()).hexdigest()[:8]
                    if with_hash
                    or len(named_args_string) > 28
                    or "'" in named_args_string
                    or "{" in named_args_string
                    else named_args_string
                )

            name_signature = (
                clean_name(f"{prefix}_{named_args_string}")
                if named_args_string
                else clean_value_name(prefix)
            )
            # filter the changed dictionary to only keep entries which have truly changed
            changed_arg_names = [carg.split("=")[0] for carg in changed_arg_list]
            changed = {k: changed[k] for k in changed_arg_names}
            name = name or name_signature

        else:
            raise ValueError('naming_style must be "default" or "updk"')

        name = get_name_short(name, max_name_length=max_name_length)
        decorator = kwargs.pop("decorator", default_decorator)
        # if no decorator is specified, but there is one specified for the active PDK, use the PDK's default decorator
        if decorator is None and active_pdk.default_decorator is not None:
            decorator = active_pdk.default_decorator

        if decorator:
            warnings.warn(
                f"decorator is deprecated and will be removed soon. {func.__name__}",
                stacklevel=2,
            )

        if name in CACHE:
            # print(f"CACHE LOAD {name} {func.__name__}({named_args_string})")
            return CACHE[name]

        # print(f"BUILD {name} {func.__name__}({named_args_string})")
        if not callable(func):
            raise ValueError(
                f"{func!r} is not callable! @cell decorator is only for functions"
            )

        if validate:
            component = validate_call(func)(*args, **kwargs)

        else:
            component = func(*args, **kwargs)

        if ports_offgrid in ("warn", "error"):
            component.assert_ports_on_grid(error_type=ports_offgrid)
        if ports_not_manhattan in ("warn", "error"):
            component.assert_ports_manhattan(error_type=ports_not_manhattan)
        if flatten:
            component = component.flatten()

        # if the component is already in the cache, but under a different alias,
        # make sure we use a copy, so we don't run into mutability errors
        if id(component) in CACHE_IDS:
            component = component.copy()

        if not isinstance(component, Component):
            raise CellReturnTypeError(
                f"function {func.__name__!r} return type = {type(component)}",
                "make sure that functions with @cell decorator return a Component",
            )

        if get_child_name:
            if not component.child:
                raise ValueError(
                    f"component {component.name} does not have a child component. "
                    "Make sure you use component_new.copy_child_info(component)"
                )
            child_name = component.child.function_name
            component_name = f"{child_name}_{name}"
            component_name = get_name_short(
                component_name, max_name_length=max_name_length
            )
        else:
            component_name = name

        if autoname:
            component.rename(component_name, max_name_length=max_name_length)
        if get_child_name:
            CACHE[name] = component

        if add_settings:
            component.settings = CellSettings(**full)
            component.function_name = func.__name__
            component.module = func.__module__
            component.__doc__ = func.__doc__

        for post in post_process or []:
            component = post(component)

        component.info.update(metadata)

        if decorator:
            if not callable(decorator):
                raise ValueError(f"decorator = {type(decorator)} needs to be callable")
            component_new = decorator(component)
            component = component_new or component
            CACHE[component_name] = component

        component._locked = True
        CACHE_IDS.add(id(component))
        return component

    sig = inspect.signature(func)
    wrapper.__signature__ = sig.replace(return_annotation=Component)  # type: ignore
    return wrapper


cell_without_validator = cell
cell_with_module = partial(cell, include_module=True)
cell_import_gds = partial(cell, autoname=False, add_settings=False)
cell_with_child = partial(cell, get_child_name=True)


@cell_with_child
def container(
    component,
    function,
    **kwargs,
) -> gf.Component:
    """Returns new component with a component reference.

    Args:
        component: to add to container.
        function: function to apply to component.
        kwargs: keyword arguments to pass to function.
    """

    import gdsfactory as gf

    component = gf.get_component(component)
    c = Component()
    cref = c << component
    c.add_ports(cref.ports)

    function(c, **kwargs)
    c.copy_child_info(component)
    return c


if __name__ == "__main__":
    from functools import partial

    import gdsfactory as gf

    c = partial(gf.components.mzi)
    c = gf.routing.add_fiber_array(c)
    c.show()

    # c = gf.components.straight(info={"simulation": "eme"}, name="hi")
    # c = gf.components.straight()
    # c = gf.Component()
    # print(type(c.info))
    # print(c.name)
    # print(c.info["simulation"])
    # c.show()
